define(function(require) {
    
'use strict';
/* globals gettext */

var _ = require('libs/underscore'),
    templates = require('plugins/core/templates'),
    footnote = require('plugins/core/footnote'),
    switchTo = require('plugins/core/switch'),
    lists = require('plugins/core/lists'),
    plugin = {name: 'core', actions: [], canvas: {}, documentExtension: {textNode: {}}},
    Dialog = require('views/dialog/dialog');


plugin.documentExtension.textNode.transformations = {
    breakContent: {
        impl: function(args) {
            var node = this,
                newNodes, emptyText;
            newNodes = node.split({offset: args.offset});
            [newNodes.first, newNodes.second].some(function(newNode) {
                if(!(newNode.contents().length)) {
                    emptyText = newNode.append({text: ''});
                    return true; // break
                }
            });
            return _.extend(newNodes, {emptyText: emptyText});
        },
        getChangeRoot: function() {
            return this.context.parent().parent();
        }
    },
    mergeContentUp: function() {
        /* globals Node */
        var myPrev = this,
            base = this,
            ret;

        if(myPrev.nodeType === Node.TEXT_NODE) {
            if(myPrev.getIndex() > 0) {
                return;
            }
            myPrev = base = myPrev.parent();
        }

        myPrev = myPrev && myPrev.prev();

        if(myPrev && myPrev.nodeType === Node.ELEMENT_NODE)  {
            var ptr = this,
                next;
            while(ptr) {
                next = ptr.next();
                if(!ret) {
                    ret = myPrev.append(ptr);
                } else {
                    myPrev.append(ptr);
                }
                
                ptr = next;
            }
            if(base !== this) {
                base.detach();
            }
            return {node: ret, offset: ret.sameNode(this) ? null : ret.getText().length - this.getText().length};
        }
    }
};

var undoRedoAction = function(dir) {
    return {
        name: dir,
        params: {
            document: {type: 'context', name: 'document'},
        },
        stateDefaults: {
            label: dir === 'undo' ? '<-' : '->',
            icon: 'share-alt',
            iconStyle: dir === 'undo' ? '-webkit-transform: scale(-1,1); transform: scale(-1, 1)' : '',
            execute: function(callback, params) {
                params.document[dir]();
                callback();
            },
        },
        getState: function(params) {
            var allowed = params.document && !!(params.document[dir+'Stack'].length),
                desc = dir === 'undo' ? gettext('Undo') : gettext('Redo'),
                descEmpty = dir === 'undo' ? gettext('There is nothing to undo') : gettext('There is nothing to redo');
            if(allowed) {
                var metadata = _.last(params.document[dir+'Stack']).metadata;
                if(metadata) {
                    desc += ': ' + (metadata.description || gettext('unknown operation'));
                }
            }
            return {
                allowed: allowed,
                description: allowed ? desc : descEmpty
            };
        }
    };
};

var pad = function(number) {
    if(number < 10) {
        number = '0' + number;
    }
    return number;
};

var commentAction = {
    name: 'comment',
    params: {
        fragment: {type: 'context', name: 'fragment'}
    },
    stateDefaults: {
        icon: 'comment',
        execute: function(callback, params, editor) {
            /* globals Node */
            var node = params.fragment.node,
                action = this;
            if(node.nodeType === Node.TEXT_NODE) {
                node = node.parent();
            }
            node.document.transaction(function() {
                var comment =  node.after({tagName: 'aside', attrs: {'class': 'comment'}});
                comment.append({text:''});
                var user = editor.getUser(), creator;
                if(user) {
                    creator = user.name;
                    if(user.email) {
                        creator += ' (' + user.email + ')';
                    }
                } else {
                    creator = 'anonymous';
                }

                var currentDate = new Date(),
                    dt = pad(currentDate.getDate()) + '-' +
                                pad((currentDate.getMonth() + 1))  + '-' +
                                pad(currentDate.getFullYear()) + ' ' +
                                pad(currentDate.getHours()) + ':' +
                                pad(currentDate.getMinutes()) + ':' +
                                pad(currentDate.getSeconds());

                var metadata = comment.getMetadata();
                metadata.add({key: 'creator', value: creator});
                metadata.add({key: 'date', value: dt});
            }, {
                metadata: {
                    description: action.getState().description
                },
                success: callback
            });
        },
    },
    getState: function(params) {
        var state = {
            allowed: params.fragment && params.fragment.isValid() &&
                        params.fragment instanceof params.fragment.NodeFragment && !params.fragment.node.isRoot()
        };
        if(state.allowed) {
            state.description = gettext('Insert comment');
        }
        return state;
    }
};


var createWrapTextAction = function(createParams) {
    return {
        name: createParams.name,
        params: {
            fragment: {type: 'context', name: 'fragment'},
        },
        getState: function(params) {
            var state = {
                    label: this.config.label
                },
                parent;
            
            if(!params.fragment || !params.fragment.isValid()) {
                return _.extend(state, {allowed: false});
            }

            if(params.fragment instanceof params.fragment.CaretFragment && params.fragment.node.isInside(createParams.klass)) {
                return _.extend(state, {
                    allowed: true,
                    toggled: true,
                    description: createParams.unwrapDescription,
                    execute: function(callback, params) {
                        var node = params.fragment.node,
                            doc = node.document,
                            toRemove = node.getParent(createParams.klass),
                            prefLen = 0;

                        if(node.sameNode(toRemove.contents()[0]) && toRemove.isPrecededByTextNode()) {
                            prefLen = toRemove.prev().getText().length;
                        }

                        doc.transaction(function() {
                            var ret = toRemove.unwrapContent(),
                                newFragment = params.fragment;
                            if(!newFragment.isValid()) {
                                newFragment =  doc.createFragment(doc.CaretFragment, {
                                    node: ret.element1,
                                    offset: prefLen + params.fragment.offset
                                });
                            }
                            return newFragment;
                        }, {
                            metadata: {
                                description: createParams.unwrapDescription
                            },
                            success: callback
                        });
                    }
                });
            }

            if(params.fragment instanceof params.fragment.TextRangeFragment && params.fragment.hasSiblingBoundries()) {
                parent = params.fragment.startNode.parent();
                if(parent && parent.is(createParams.klass) || parent.isInside(createParams.klass)) {
                    return _.extend(state, {allowed: false});
                }

                return _.extend(state, {
                    allowed: true,
                    description: createParams.wrapDescription,
                    execute: function(callback, params) {
                        params.fragment.document.transaction(function() {
                            var parent = params.fragment.startNode.parent(),
                                doc = params.fragment.document,
                                wrapper, lastTextNode;
                            
                            wrapper = parent.wrapText({
                                _with: {tagName: 'span', attrs: {'class': createParams.klass}},
                                offsetStart: params.fragment.startOffset,
                                offsetEnd: params.fragment.endOffset,
                                textNodeIdx: [params.fragment.startNode.getIndex(), params.fragment.endNode.getIndex()]
                            });
                                
                            lastTextNode = wrapper.getLastTextNode();
                            if(lastTextNode) {
                                return doc.createFragment(doc.CaretFragment, {node: lastTextNode, offset: lastTextNode.getText().length});
                            }
                        }, {
                            metadata: {
                                description: createParams.wrapDescription
                            },
                            success: callback
                        });
                    }
                });
            }
            return _.extend(state, {allowed: false});
        }
    };
};


var createLinkFromSelection = function(callback, params) {
    var doc = params.fragment.document,
        dialog = Dialog.create({
            title: gettext('Create link'),
            executeButtonText: gettext('Apply'),
            cancelButtonText: gettext('Cancel'),
            fields: [
                {label: gettext('Link'), name: 'href', type: 'input'}
            ]
        }),
        action = this;
    
    dialog.on('execute', function(event) {
        doc.transaction(function() {
            var span =  action.params.fragment.startNode.parent().wrapText({
                _with: {tagName: 'span', attrs: {'class': 'link'}},
                offsetStart: params.fragment.startOffset,
                offsetEnd: params.fragment.endOffset,
                textNodeIdx: [params.fragment.startNode.getIndex(), params.fragment.endNode.getIndex()]
            }),
                doc = params.fragment.document;
            span.setAttr('href', event.formData.href);
            event.success();
            return doc.createFragment(doc.CaretFragment, {node: span.contents()[0], offset:0});
        }, {
            metadata: {
                description: action.getState().description
            },
            success: callback
        });
    });
    dialog.show();
};

var editLink = function(callback, params) {
    var doc = params.fragment.document,
        link = params.fragment.node.getParent('link'),
        dialog = Dialog.create({
            title: gettext('Edit link'),
            executeButtonText: gettext('Apply'),
            cancelButtonText: gettext('Cancel'),
            fields: [
                {label: gettext('Link'), name: 'href', type: 'input', initialValue: link.getAttr('href')}
            ]
        }),
        action = this;
    
    dialog.on('execute', function(event) {
        doc.transaction(function() {
            link.setAttr('href', event.formData.href);
            event.success();
            return params.fragment;
        }, {
            metadata: {
                description: action.getState().description
            },
            success: callback
        });
    });
    dialog.show();
};

var linkAction = {
    name: 'link',
    params: {
        fragment: {type: 'context', name: 'fragment'}
    },
    stateDefaults: {
        label: gettext('link')
    },
    getState: function(params) {
        if(!params.fragment || !params.fragment.isValid()) {
            return {allowed: false};
        }

        if(params.fragment instanceof params.fragment.TextRangeFragment) {
            if(!params.fragment.hasSiblingBoundries() || params.fragment.startNode.parent().is('link')) {
                return {allowed: false};
            }
            return {
                allowed: true,
                description: gettext('Create link from selection'),
                execute: createLinkFromSelection
            };
        }

        if(params.fragment instanceof params.fragment.CaretFragment) {
            if(params.fragment.node.isInside('link')) {
                return {
                    allowed: true,
                    toggled: true,
                    description: gettext('Edit link'),
                    execute: editLink
                };
            }
        }
        return {allowed: false};
    }
};


plugin.actions = [
    undoRedoAction('undo'),
    undoRedoAction('redo'),
    commentAction,
    createWrapTextAction({name: 'emphasis', klass: 'emp', wrapDescription: gettext('Mark as emphasized'), unwrapDescription: gettext('Remove emphasis')}),
    createWrapTextAction({name: 'cite', klass: 'cite', wrapDescription: gettext('Mark as citation'), unwrapDescription: gettext('Remove citation')}),
    linkAction
].concat(plugin.actions, templates.actions, footnote.actions, switchTo.actions, lists.actions);



plugin.config = function(config) {
    // templates.actions[0].config(config.templates);
    templates.actions[0].params.template.options = config.templates;
};

return plugin;

});